import 'dart:core';

import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:equatable/equatable.dart';
import 'package:flutter/foundation.dart';
part 'crud_repository.dart';
part 'exceptions.dart';
part 'query_filter.dart';
part 'firebase_model.dart';

class FirestoreRepository extends CRUDRepository {
  FirestoreRepository({FirebaseFirestore? firestore, bool persistenceEnabled = false})
      : this.firestore = firestore ?? FirebaseFirestore.instance {
    if (kIsWeb) {
      if (persistenceEnabled) this.firestore.enablePersistence();
    } else
      this.firestore.settings = Settings(persistenceEnabled: persistenceEnabled);
  }

  late final FirebaseFirestore firestore;

  DocumentReference<Map<String, dynamic>> doc(String path) => firestore.doc(path);

  Future<QuerySnapshot<Map<String, dynamic>>> getCollection(List<String> fields,
      {List<QueryFilter>? filters,
      DocumentSnapshot<Map<String, dynamic>>? startAfter,
      int? limit,
      GetOptions? getOptions}) async {
    return getQuery(fields, filters: filters, startAfter: startAfter, limit: limit).get(getOptions);
  }

  Query<Map<String, Object?>> getQuery(List<String> fields,
      {List<QueryFilter>? filters, DocumentSnapshot<Map<String, dynamic>>? startAfter, int? limit}) {
    if (fields.contains(null)) throw FirestoreNullArgumentException();
    if (fields.length % 2 == 0) throw FirestoreArgumentException();
    Query<Map<String, Object?>> query = firestore.collection(fields.join('/'));
    if (filters != null) {
      for (var filter in filters)
        query = query.where(
          filter.field,
          isEqualTo: filter.isEqualTo,
          isLessThan: filter.isLessThan,
          isLessThanOrEqualTo: filter.isLessThanOrEqualTo,
          isGreaterThan: filter.isGreaterThan,
          isGreaterThanOrEqualTo: filter.isGreaterThanOrEqualTo,
          arrayContains: filter.arrayContains,
          arrayContainsAny: filter.arrayContainsAny,
          whereIn: filter.whereIn,
          isNull: filter.isNull,
        );
    }

    if (startAfter != null) query = query.startAfterDocument(startAfter);
    if (limit != null) query = query.limit(limit);

    return limitQuery(query, startAfter: startAfter, limit: limit);
  }

  Query<Map<String, Object?>> limitQuery(Query<Map<String, Object?>> query,
      {DocumentSnapshot<Map<String, dynamic>>? startAfter, int? limit}) {
    if (startAfter != null) query = query.startAfterDocument(startAfter);
    if (limit != null) query = query.limit(limit);

    return query;
  }

  Future<DocumentSnapshot<Map<String, dynamic>>> getDocument(List<String> fields) {
    if (fields.contains(null)) throw FirestoreNullArgumentException();
    if (fields.length % 2 != 0) throw FirestoreArgumentException();
    return read(fields.join('/'));
  }

  Future<void> addData(List<String> fields, Map<String, Object?> jsonData) {
    if (fields.contains(null)) throw FirestoreNullArgumentException();
    if (fields.length % 2 == 0) throw FirestoreArgumentException();
    return create(fields.join('/'), jsonData);
  }

  Future<void> setData(List<String> fields, Map<String, Object?> jsonData) {
    if (fields.contains(null)) throw FirestoreNullArgumentException();
    if (fields.length % 2 != 0) throw FirestoreArgumentException();
    return _set(fields.join('/'), jsonData);
  }

  Future<void> updateData(List<String> fields, Map<String, Object?> jsonData) {
    if (fields.contains(null)) throw FirestoreNullArgumentException();
    if (fields.length % 2 != 0) throw FirestoreArgumentException();
    return update(fields.join('/'), jsonData);
  }

  Stream<DocumentSnapshot<Map<String, dynamic>>> listen(List<String> fields) {
    if (fields.contains(null)) throw FirestoreNullArgumentException();
    if (fields.length % 2 != 0) throw FirestoreArgumentException();
    return firestore.doc(fields.join('/')).snapshots();
  }

  @override
  Future<DocumentReference<Map<String, dynamic>>> create(String path, Map<String, dynamic> data) {
    return firestore.collection(path).add(data);
  }

  @override
  Future<void> delete(String path) {
    return firestore.doc(path).delete();
  }

  @override
  Future<DocumentSnapshot<Map<String, dynamic>>> read(String path) {
    return firestore.doc(path).get();
  }

  @override
  Future<void> update(String path, Map<String, dynamic> data) {
    return firestore.doc(path).update(data);
  }

  Future<void> _set(String path, Map<String, Object?> data) {
    return firestore.doc(path).set(data);
  }
}
