import 'dart:async';

import 'package:bloc/bloc.dart';
import 'package:equatable/equatable.dart';
import 'package:firestore_repository/firestore_repository.dart';

part 'pagination_event.dart';
part 'pagination_state.dart';

class PaginationBloc<T extends Object?> extends Bloc<PaginationEvent, PaginationState> {
  PaginationBloc(this.query, {required this.converter, this.limit = 20}) : super(PaginationInitial()) {
    on<PaginationRefreshEvent>(_onPaginationRefresh);
    on<PaginationFetched>(_onPaginationFetched);
    on<PaginationModified<T>>(_onPaginationModified);
    // on<PaginationAdded<T>>(_onPaginationModified);
    // on<PaginationRemoved<T>>(_onPaginationModified);
    _addListener();
  }

  Converter<T> converter;
  int limit;
  Query<Map<String, Object?>> query;

  late final StreamSubscription<QuerySnapshot> _listener;

  @override
  Future<void> close() {
    _listener.cancel();
    return super.close();
  }

  void _onPaginationRefresh(PaginationRefreshEvent event, Emitter<PaginationState> emit) {
    if (!(state is PaginationSuccess)) return;
    var currentState = state as PaginationSuccess<T>;
    emit(PaginationUpdate<T>.fromSuccess(currentState));
  }

  void _onPaginationFetched(PaginationFetched event, Emitter<PaginationState> emit) async {
    if (!_hasReachedMax(state)) {
      try {
        if (state is PaginationInitial) {
          var rawdata = event.documents;
          if (rawdata == null) rawdata = await _fetchPaginations(null, limit);
          final last = rawdata.isNotEmpty ? rawdata.last : null;

          final data = <T>[];
          for (var raw in rawdata) {
            var converted = await converter(raw);
            data.add(converted);
          }
          emit(PaginationSuccess<T>(data: data, hasReachedMax: data.length < limit, lastSnapshot: last));
        } else if (state is PaginationSuccess<T>) {
          var currentState = state as PaginationSuccess<T>;
          final rawdata = await _fetchPaginations(currentState.lastSnapshot, limit);
          final last = rawdata.isNotEmpty ? rawdata.last : currentState.lastSnapshot;

          final data = <T>[];
          for (var raw in rawdata) {
            var converted = await converter(raw);
            data.add(converted);
          }
          var newState = data.isEmpty
              ? currentState.copyWith(hasReachedMax: true, lastSnapshot: last)
              : PaginationSuccess<T>(
                  data: currentState.data + data,
                  lastSnapshot: last,
                  hasReachedMax: false,
                );
          emit(newState);
        }
      } on Exception catch (e) {
        emit(PaginationFailure(e.toString()));
      }
    }
  }

  void _onPaginationModified(PaginationModified<T> event, Emitter<PaginationState> emit) async {
    if (!(state is PaginationSuccess)) return;
    var currentState = state as PaginationSuccess<T>;
    var converted = event.document;
    if (event is PaginationAdded) {
      currentState.data.insert(0, converted);
    } else if (event is PaginationRemoved) {
      currentState.data.remove(converted);
    } else {
      var index = currentState.data.indexWhere((element) => element == converted);
      if (index >= 0) {
        print(index);
        currentState.data[index] = converted;
      }
    }
    emit(PaginationUpdate<T>.fromSuccess(currentState));
  }

  void _addListener() {
    // if (_listener != null) return;
    _listener = query.snapshots().listen((event) async {
      // return;
      if (state is PaginationInitial) {
        add(PaginationFetched(documents: event.docChanges.map((e) => e.doc).toList()));
      } else {
        for (var change in event.docChanges) {
          var converted = await converter(change.doc);
          if (change.type == DocumentChangeType.added)
            add(PaginationAdded<T>(converted));
          else if (change.type == DocumentChangeType.modified) {
            add(PaginationModified<T>(converted));
          } else if (change.type == DocumentChangeType.removed) {
            add(PaginationRemoved<T>(converted));
          }
        }
      }
    });
  }

  bool _hasReachedMax(PaginationState state) => state is PaginationSuccess && state.hasReachedMax;

  Future<List<DocumentSnapshot<Map<String, dynamic>>>> _fetchPaginations(
      DocumentSnapshot<Map<String, dynamic>>? startAfter, int limit) async {
    var q = query;
    if (startAfter != null) q = q.startAfterDocument(startAfter);

    var snapshot = await q.limit(limit).get();
    // .catchError(() {
    //   throw Exception('error fetching data');
    // });

    return snapshot.docs;
  }
}
